"""Tests for the owlbear MCP server."""

import json
import pytest
import polars as pl
from unittest.mock import patch, MagicMock

from owlbear.mcp_server import (
    _get_client,
    _query_to_json,
    _is_scalar_stat_type,
    _MAX_ROWS_CAP,
    execute_query,
    list_databases,
    list_tables,
    describe_table,
    get_schema_context,
    profile_table,
    generate_snippet,
    explore_table,
    build_pipeline,
    get_config,
)
import owlbear.mcp_server as mcp_mod


@pytest.fixture(autouse=True)
def _reset_client():
    """Reset the lazy singleton before each test."""
    mcp_mod._client = None
    yield
    mcp_mod._client = None


# ---------------------------------------------------------------------------
# Client factory
# ---------------------------------------------------------------------------


class TestGetClient:
    def test_athena_default(self):
        env = {
            "OWLBEAR_S3_OUTPUT_LOCATION": "s3://bucket/results/",
            "OWLBEAR_DATABASE": "mydb",
        }
        with patch.dict("os.environ", env, clear=False), patch(
            "owlbear.mcp_server.AthenaClient"
        ) as MockAthena:
            client = _get_client()
            MockAthena.assert_called_once_with(
                database="mydb",
                output_location="s3://bucket/results/",
                region="us-east-1",
            )
            assert client is MockAthena.return_value

    def test_athena_explicit(self):
        env = {
            "OWLBEAR_BACKEND": "athena",
            "OWLBEAR_S3_OUTPUT_LOCATION": "s3://b/r/",
            "AWS_REGION": "eu-west-1",
        }
        with patch.dict("os.environ", env, clear=False), patch(
            "owlbear.mcp_server.AthenaClient"
        ) as MockAthena:
            client = _get_client()
            MockAthena.assert_called_once_with(
                database="default",
                output_location="s3://b/r/",
                region="eu-west-1",
            )
            assert client is MockAthena.return_value

    def test_trino_backend(self):
        env = {
            "OWLBEAR_BACKEND": "trino",
            "OWLBEAR_TRINO_HOST": "trino.example.com",
            "OWLBEAR_TRINO_PORT": "8080",
            "OWLBEAR_TRINO_USER": "analyst",
            "OWLBEAR_TRINO_CATALOG": "hive",
            "OWLBEAR_DATABASE": "default",
        }
        with patch.dict("os.environ", env, clear=False), patch(
            "owlbear.mcp_server.TrinoClient"
        ) as MockTrino:
            client = _get_client()
            MockTrino.assert_called_once_with(
                host="trino.example.com",
                port=8080,
                user="analyst",
                catalog="hive",
                schema="default",
            )
            assert client is MockTrino.return_value

    def test_unknown_backend(self):
        env = {"OWLBEAR_BACKEND": "spark"}
        with patch.dict("os.environ", env, clear=False):
            with pytest.raises(ValueError, match="Unknown OWLBEAR_BACKEND"):
                _get_client()

    def test_lazy_singleton(self):
        """Second call returns the same object without re-creating."""
        sentinel = object()
        mcp_mod._client = sentinel
        assert _get_client() is sentinel


# ---------------------------------------------------------------------------
# _query_to_json helper
# ---------------------------------------------------------------------------


class TestQueryToJson:
    def test_athena_path(self):
        mock_client = MagicMock(spec_set=["query", "results"])
        # Make isinstance(client, AthenaClient) return True
        with patch("owlbear.mcp_server._get_client", return_value=mock_client), patch(
            "owlbear.mcp_server.isinstance", side_effect=lambda obj, cls: cls.__name__ == "AthenaClient"
        ):
            # Simpler approach: just set the class
            pass

        # Use a real approach: patch _get_client and check behaviour
        from owlbear.athena import AthenaClient

        mock_athena = MagicMock(spec=AthenaClient)
        mock_athena.query.return_value = "eid-123"
        mock_athena.results.return_value = pl.DataFrame({"x": [1, 2]})

        with patch("owlbear.mcp_server._get_client", return_value=mock_athena):
            result = _query_to_json("SELECT 1", 100)

        mock_athena.query.assert_called_once_with("SELECT 1")
        mock_athena.results.assert_called_once_with("eid-123", max_rows=100)
        assert json.loads(result) == [{"x": 1}, {"x": 2}]

    def test_trino_path(self):
        from owlbear.trino import TrinoClient

        mock_trino = MagicMock(spec=TrinoClient)
        mock_trino.query.return_value = pl.DataFrame({"y": [3]})

        with patch("owlbear.mcp_server._get_client", return_value=mock_trino):
            result = _query_to_json("SELECT 1", 50)

        mock_trino.query.assert_called_once_with("SELECT 1", max_rows=50)
        assert json.loads(result) == [{"y": 3}]

    def test_max_rows_capped(self):
        from owlbear.athena import AthenaClient

        mock_athena = MagicMock(spec=AthenaClient)
        mock_athena.query.return_value = "eid"
        mock_athena.results.return_value = pl.DataFrame({"x": [1]})

        with patch("owlbear.mcp_server._get_client", return_value=mock_athena):
            _query_to_json("SELECT 1", 999_999)

        mock_athena.results.assert_called_once_with("eid", max_rows=_MAX_ROWS_CAP)


# ---------------------------------------------------------------------------
# MCP tools
# ---------------------------------------------------------------------------


class TestExecuteQuery:
    def test_success(self):
        with patch(
            "owlbear.mcp_server._query_to_json", return_value='[{"a":1}]'
        ) as mock_q:
            result = execute_query("SELECT 1", max_rows=10)
            mock_q.assert_called_once_with("SELECT 1", 10)
            assert result == '[{"a":1}]'

    def test_default_max_rows(self):
        with patch(
            "owlbear.mcp_server._query_to_json", return_value="[]"
        ) as mock_q:
            execute_query("SELECT 1")
            mock_q.assert_called_once_with("SELECT 1", 500)

    def test_error_returns_json(self):
        with patch(
            "owlbear.mcp_server._query_to_json", side_effect=RuntimeError("boom")
        ):
            result = execute_query("BAD SQL")
            parsed = json.loads(result)
            assert "error" in parsed
            assert "boom" in parsed["error"]


class TestListDatabases:
    def test_success(self):
        with patch(
            "owlbear.mcp_server._query_to_json", return_value='[{"database":"db1"}]'
        ) as mock_q:
            result = list_databases()
            mock_q.assert_called_once_with("SHOW DATABASES", max_rows=_MAX_ROWS_CAP)
            assert json.loads(result) == [{"database": "db1"}]

    def test_error_returns_json(self):
        with patch(
            "owlbear.mcp_server._query_to_json", side_effect=Exception("fail")
        ):
            result = list_databases()
            assert "error" in json.loads(result)


class TestListTables:
    def test_no_database(self):
        with patch(
            "owlbear.mcp_server._query_to_json", return_value="[]"
        ) as mock_q:
            list_tables()
            mock_q.assert_called_once_with("SHOW TABLES", max_rows=_MAX_ROWS_CAP)

    def test_with_database(self):
        with patch(
            "owlbear.mcp_server._query_to_json", return_value="[]"
        ) as mock_q:
            list_tables(database="analytics")
            mock_q.assert_called_once_with(
                "SHOW TABLES IN analytics", max_rows=_MAX_ROWS_CAP
            )

    def test_error_returns_json(self):
        with patch(
            "owlbear.mcp_server._query_to_json", side_effect=Exception("fail")
        ):
            result = list_tables()
            assert "error" in json.loads(result)


class TestDescribeTable:
    def test_success(self):
        with patch(
            "owlbear.mcp_server._query_to_json", return_value="[]"
        ) as mock_q:
            describe_table("my_db.my_table")
            mock_q.assert_called_once_with(
                "DESCRIBE my_db.my_table", max_rows=_MAX_ROWS_CAP
            )

    def test_error_returns_json(self):
        with patch(
            "owlbear.mcp_server._query_to_json", side_effect=Exception("fail")
        ):
            result = describe_table("bad_table")
            assert "error" in json.loads(result)


# ---------------------------------------------------------------------------
# _is_scalar_stat_type helper
# ---------------------------------------------------------------------------


class TestIsScalarStatType:
    @pytest.mark.parametrize("dtype", [
        "boolean", "tinyint", "smallint", "int", "integer", "bigint",
        "float", "double", "real",
        "varchar", "varchar(255)", "char", "string",
        "date", "timestamp", "timestamp with time zone",
        "decimal", "decimal(10,2)",
    ])
    def test_scalar_types(self, dtype: str):
        assert _is_scalar_stat_type(dtype) is True

    @pytest.mark.parametrize("dtype", [
        "array<string>", "map<string,int>", "row(x int, y int)",
        "struct<a:int>",
    ])
    def test_complex_types(self, dtype: str):
        assert _is_scalar_stat_type(dtype) is False

    def test_case_insensitive(self):
        assert _is_scalar_stat_type("VARCHAR") is True
        assert _is_scalar_stat_type("BIGINT") is True


# ---------------------------------------------------------------------------
# profile_table tool
# ---------------------------------------------------------------------------


class TestProfileTable:
    def test_success(self):
        describe_resp = json.dumps([
            {"col_name": "id", "data_type": "bigint"},
            {"col_name": "name", "data_type": "varchar"},
            {"col_name": "tags", "data_type": "array<string>"},
            {"col_name": "# Partition Information", "data_type": ""},
        ])
        count_resp = json.dumps([{"cnt": 42}])
        stats_resp = json.dumps([{
            "id__null_count": 0,
            "id__distinct_count": 42,
            "id__min": "1",
            "id__max": "42",
            "name__null_count": 2,
            "name__distinct_count": 30,
            "name__min": "alice",
            "name__max": "zed",
            "tags__null_count": 5,
        }])
        sample_resp = json.dumps([{"id": 1, "name": "alice", "tags": ["a"]}])

        with patch(
            "owlbear.mcp_server._query_to_json",
            side_effect=[describe_resp, count_resp, stats_resp, sample_resp],
        ):
            result = json.loads(profile_table("db.events", sample_size=10))

        assert result["table"] == "db.events"
        assert result["row_count"] == 42
        assert len(result["columns"]) == 3  # partition row filtered
        # scalar columns have distinct/min/max
        id_col = result["columns"][0]
        assert id_col["column"] == "id"
        assert "distinct_count" in id_col
        assert "min" in id_col
        # complex column has only null_count
        tags_col = result["columns"][2]
        assert tags_col["column"] == "tags"
        assert "distinct_count" not in tags_col
        assert len(result["sample_rows"]) == 1

    def test_error(self):
        with patch(
            "owlbear.mcp_server._query_to_json",
            side_effect=RuntimeError("timeout"),
        ):
            result = json.loads(profile_table("bad_table"))
            assert "error" in result


# ---------------------------------------------------------------------------
# get_schema_context tool
# ---------------------------------------------------------------------------


class TestGetSchemaContext:
    def test_multiple_tables(self):
        t1_resp = json.dumps([{"col_name": "a", "data_type": "int"}])
        t2_resp = json.dumps([{"col_name": "b", "data_type": "varchar"}])

        with patch(
            "owlbear.mcp_server._query_to_json",
            side_effect=[t1_resp, t2_resp],
        ):
            result = json.loads(get_schema_context("db.t1, db.t2"))

        assert "db.t1" in result
        assert "db.t2" in result
        assert result["db.t1"] == [{"col_name": "a", "data_type": "int"}]
        assert result["db.t2"] == [{"col_name": "b", "data_type": "varchar"}]

    def test_partial_failure(self):
        t1_resp = json.dumps([{"col_name": "a", "data_type": "int"}])

        with patch(
            "owlbear.mcp_server._query_to_json",
            side_effect=[t1_resp, RuntimeError("no such table")],
        ):
            result = json.loads(get_schema_context("db.t1, db.missing"))

        assert isinstance(result["db.t1"], list)
        assert "error" in result["db.missing"]

    def test_empty_input(self):
        result = json.loads(get_schema_context(""))
        assert result == {}

    def test_filters_partition_rows(self):
        resp = json.dumps([
            {"col_name": "id", "data_type": "int"},
            {"col_name": "# Partition Information", "data_type": ""},
        ])
        with patch("owlbear.mcp_server._query_to_json", return_value=resp):
            result = json.loads(get_schema_context("db.t1"))
        assert len(result["db.t1"]) == 1


# ---------------------------------------------------------------------------
# generate_snippet tool
# ---------------------------------------------------------------------------


class TestGenerateSnippet:
    def _mock_describe(self):
        return json.dumps([
            {"col_name": "user_id", "data_type": "bigint"},
            {"col_name": "name", "data_type": "varchar"},
            {"col_name": "score", "data_type": "double"},
        ])

    @pytest.mark.parametrize("operation", ["load", "filter", "aggregate", "join"])
    def test_operations(self, operation: str):
        with patch(
            "owlbear.mcp_server._query_to_json", return_value=self._mock_describe()
        ):
            result = json.loads(generate_snippet("db.users", operation))
        assert result["table"] == "db.users"
        assert result["operation"] == operation
        assert "AthenaClient" in result["snippet"]
        assert "db.users" in result["snippet"]

    def test_unknown_operation(self):
        result = json.loads(generate_snippet("db.users", "pivot"))
        assert "error" in result
        assert "pivot" in result["error"]

    def test_error_during_describe(self):
        with patch(
            "owlbear.mcp_server._query_to_json",
            side_effect=RuntimeError("access denied"),
        ):
            result = json.loads(generate_snippet("db.users", "load"))
        assert "error" in result


# ---------------------------------------------------------------------------
# MCP prompts
# ---------------------------------------------------------------------------


class TestExploreTablePrompt:
    def test_returns_message_list(self):
        result = explore_table("db.events")
        assert isinstance(result, list)
        assert len(result) == 1
        msg = result[0]
        assert msg["role"] == "user"
        assert "db.events" in msg["content"]
        assert "DESCRIBE" in msg["content"]

    def test_table_name_in_content(self):
        result = explore_table("analytics.clicks")
        assert "analytics.clicks" in result[0]["content"]


class TestBuildPipelinePrompt:
    def test_returns_message_list(self):
        result = build_pipeline("db.orders", "compute daily revenue")
        assert isinstance(result, list)
        assert len(result) == 1
        msg = result[0]
        assert msg["role"] == "user"
        assert "db.orders" in msg["content"]
        assert "compute daily revenue" in msg["content"]

    def test_includes_boilerplate(self):
        result = build_pipeline("db.t", "test")
        content = result[0]["content"]
        assert "AthenaClient" in content
        assert "import polars" in content


# ---------------------------------------------------------------------------
# MCP resource: owlbear://config
# ---------------------------------------------------------------------------


class TestGetConfigResource:
    def test_athena_config(self):
        env = {
            "OWLBEAR_DATABASE": "mydb",
            "AWS_REGION": "us-west-2",
            "OWLBEAR_S3_OUTPUT_LOCATION": "s3://bucket/out/",
        }
        with patch.dict("os.environ", env, clear=False):
            result = json.loads(get_config())
        assert result["backend"] == "athena"
        assert result["database"] == "mydb"
        assert result["region"] == "us-west-2"
        assert result["s3_output_location"] == "s3://bucket/out/"

    def test_trino_config(self):
        env = {
            "OWLBEAR_BACKEND": "trino",
            "OWLBEAR_TRINO_HOST": "trino.example.com",
            "OWLBEAR_TRINO_PORT": "8080",
            "OWLBEAR_TRINO_USER": "analyst",
            "OWLBEAR_TRINO_CATALOG": "hive",
            "OWLBEAR_DATABASE": "default",
        }
        with patch.dict("os.environ", env, clear=False):
            result = json.loads(get_config())
        assert result["backend"] == "trino"
        assert result["host"] == "trino.example.com"
        assert result["port"] == 8080
        assert result["user"] == "analyst"
        assert result["catalog"] == "hive"
        assert result["database"] == "default"



#!/usr/bin/env python3
"""TrinoClient — execute Trino SQL, get typed Polars DataFrames."""

import polars as pl
import pyarrow as pa
from collections.abc import Sequence
from typing import Iterator, Optional, Any, cast

from .types import presto_type_to_pyarrow


class TrinoClient:
    def __init__(
        self,
        host: str,
        port: int = 443,
        user: Optional[str] = None,
        catalog: Optional[str] = None,
        schema: Optional[str] = None,
        auth: Optional[Any] = None,
        http_scheme: str = "https",
        **kwargs,
    ):
        self.host = host
        self.port = port
        self.user = user
        self.catalog = catalog
        self.schema = schema
        self.auth = auth
        self.http_scheme = http_scheme
        self.kwargs = kwargs

    def _get_connection(self) -> Any:
        """Create a new Trino connection."""
        from trino.dbapi import connect

        params = {
            "host": self.host,
            "port": self.port,
            "http_scheme": self.http_scheme,
        }
        if self.user is not None:
            params["user"] = self.user
        if self.catalog is not None:
            params["catalog"] = self.catalog
        if self.schema is not None:
            params["schema"] = self.schema
        if self.auth is not None:
            params["auth"] = self.auth
        params.update(self.kwargs)

        return connect(**params)

    def query(
        self,
        query: str,
        max_rows: int = 0,
        parameters: Optional[Sequence[Any]] = None,
    ) -> pl.DataFrame:
        """Execute a query and return results as a Polars DataFrame.

        Args:
            query: SQL query string.
            max_rows: Maximum rows to return. 0 means no limit.
            parameters: Query parameters passed to ``cursor.execute``.

        Returns:
            A Polars DataFrame with typed columns.
        """
        conn = self._get_connection()
        try:
            cursor = conn.cursor()
            cursor.execute(query, params=parameters)

            # Fetch rows
            if max_rows > 0:
                rows = cursor.fetchmany(max_rows)
            else:
                rows = cursor.fetchall()

            description = cursor.description
            if not description:
                return pl.DataFrame()

            col_names = [desc[0] for desc in description]
            col_types = [desc[1] for desc in description]

            if not rows:
                # Build empty DataFrame with correct schema
                schema_fields = []
                for name, type_str in zip(col_names, col_types):
                    pa_type = (
                        presto_type_to_pyarrow(type_str) if type_str else pa.string()
                    )
                    schema_fields.append(pa.field(name, pa_type))
                arrow_schema = pa.schema(schema_fields)
                return cast(
                    pl.DataFrame,
                    pl.from_arrow(
                        pa.table({name: [] for name in col_names}, schema=arrow_schema)
                    ),
                )

            # Transpose rows into columns
            columns_data = list(zip(*rows))

            # Build PyArrow arrays
            arrays = []
            fields = []
            for i, (name, type_str) in enumerate(zip(col_names, col_types)):
                pa_type = presto_type_to_pyarrow(type_str) if type_str else pa.string()
                data = list(columns_data[i])
                try:
                    array = pa.array(data, type=pa_type)
                except (pa.ArrowInvalid, pa.ArrowTypeError):
                    array = pa.array(
                        [str(x) if x is not None else None for x in data],
                        type=pa.string(),
                    )
                    pa_type = pa.string()
                arrays.append(array)
                fields.append(pa.field(name, pa_type))

            table = pa.table(arrays, names=col_names)
            return cast(pl.DataFrame, pl.from_arrow(table))
        finally:
            conn.close()

    def query_iter(
        self,
        query: str,
        page_size: int = 1000,
        parameters: Optional[Sequence[Any]] = None,
    ) -> Iterator[pl.DataFrame]:
        """Execute a query and yield results page-by-page as Polars DataFrames.

        Args:
            query: SQL query string.
            page_size: Rows per ``fetchmany`` call.
            parameters: Query parameters passed to ``cursor.execute``.

        Yields:
            One ``pl.DataFrame`` per batch.
        """
        conn = self._get_connection()
        try:
            cursor = conn.cursor()
            cursor.execute(query, params=parameters)

            description = cursor.description
            if not description:
                return

            col_names = [desc[0] for desc in description]
            col_types = [desc[1] for desc in description]

            # Build PyArrow schema (reused for every page)
            schema_fields = []
            for name, type_str in zip(col_names, col_types):
                pa_type = presto_type_to_pyarrow(type_str) if type_str else pa.string()
                schema_fields.append(pa.field(name, pa_type))
            arrow_schema = pa.schema(schema_fields)

            while True:
                rows = cursor.fetchmany(page_size)
                if not rows:
                    break

                columns_data = list(zip(*rows))

                arrays = []
                for i, (name, type_str) in enumerate(zip(col_names, col_types)):
                    pa_type = presto_type_to_pyarrow(type_str) if type_str else pa.string()
                    data = list(columns_data[i])
                    try:
                        array = pa.array(data, type=pa_type)
                    except (pa.ArrowInvalid, pa.ArrowTypeError):
                        array = pa.array(
                            [str(x) if x is not None else None for x in data],
                            type=pa.string(),
                        )
                    arrays.append(array)

                table = pa.table(arrays, names=col_names)
                yield cast(pl.DataFrame, pl.from_arrow(table))
        finally:
            conn.close()



#!/usr/bin/env python3
"""AthenaClient — execute Athena SQL, get typed Polars DataFrames."""

import boto3
import polars as pl
import pyarrow as pa
import time
from typing import Iterator, Optional, Any, cast
from botocore.config import Config

from .types import presto_type_to_pyarrow


class AthenaClient:
    def __init__(
        self,
        database: str,
        output_location: str,
        region: str = "us-east-1",
        session: Optional[boto3.Session] = None,
        config: Optional[Config] = None,
        **client_kwargs,
    ):
        # Use provided session or create new one
        if session:
            self.client = session.client("athena", config=config, **client_kwargs)
        else:
            # Default config with retries
            default_config = Config(
                region_name=region,
                retries={"max_attempts": 3, "mode": "adaptive"},
                max_pool_connections=50,
            )
            final_config = config or default_config
            self.client = boto3.client("athena", config=final_config, **client_kwargs)

        self.database = database
        self.output_location = output_location

    def query(
        self,
        query: str,
        wait_for_completion: bool = True,
        work_group: Optional[str] = None,
        query_context: Optional[dict[str, str]] = None,
        result_config: Optional[dict[str, Any]] = None,
        parameters: Optional[list[str]] = None,
        result_reuse_max_age: Optional[int] = None,
    ) -> str:
        """Execute a query and return the execution ID.

        Args:
            query: SQL query string.
            wait_for_completion: Block until the query finishes.
            work_group: Athena workgroup name.
            query_context: Extra context key/values (merged with Database).
            result_config: Extra result-configuration key/values.
            parameters: Parameterized query execution parameters.
            result_reuse_max_age: If set, enable result reuse for this many
                minutes via ``ResultReuseConfiguration``.
        """
        # Build query execution context
        context = {"Database": self.database}
        if query_context:
            context.update(query_context)

        # Build result configuration
        config = {"OutputLocation": self.output_location}
        if result_config:
            config.update(result_config)

        # Prepare execution parameters
        params: dict[str, Any] = {
            "QueryString": query,
            "QueryExecutionContext": context,
            "ResultConfiguration": config,
        }

        if work_group:
            params["WorkGroup"] = work_group

        if parameters is not None:
            params["ExecutionParameters"] = parameters

        if result_reuse_max_age is not None:
            params["ResultReuseConfiguration"] = {
                "ResultReuseByAgeConfiguration": {
                    "Enabled": True,
                    "MaxAgeInMinutes": result_reuse_max_age,
                }
            }

        try:
            response = self.client.start_query_execution(**params)
            execution_id = response["QueryExecutionId"]

            if wait_for_completion:
                self._wait_for_completion(execution_id)

            return execution_id

        except Exception as e:
            raise Exception(f"Failed to execute query: {e}") from e

    def _wait_for_completion(self, execution_id: str, max_wait_time: int = 300) -> str:
        """Wait for query to complete with exponential backoff"""
        start_time = time.time()
        sleep_time = 1

        while time.time() - start_time < max_wait_time:
            try:
                response = self.client.get_query_execution(
                    QueryExecutionId=execution_id
                )
                status = response["QueryExecution"]["Status"]["State"]

                if status in ["SUCCEEDED", "FAILED", "CANCELLED"]:
                    if status == "FAILED":
                        reason = response["QueryExecution"]["Status"].get(
                            "StateChangeReason", "Unknown error"
                        )
                        error_msg = (
                            response["QueryExecution"]["Status"]
                            .get("AthenaError", {})
                            .get("ErrorMessage", "")
                        )
                        raise Exception(f"Query failed: {reason}. Error: {error_msg}")
                    elif status == "CANCELLED":
                        raise Exception("Query was cancelled")
                    return status

                time.sleep(sleep_time)
                sleep_time = min(sleep_time * 1.5, 10)  # Exponential backoff, max 10s

            except Exception as e:
                # Only handle specific AWS exceptions, let our own exceptions through
                if "Query failed:" in str(e) or "Query was cancelled" in str(e):
                    raise  # Re-raise our own exceptions
                elif "InvalidRequestException" in str(
                    type(e)
                ) or "InvalidRequestException" in str(e):
                    raise Exception(
                        f"Invalid query execution ID: {execution_id}"
                    ) from e
                else:
                    # For any other exception, continue the loop or re-raise
                    break

        raise TimeoutError(f"Query did not complete within {max_wait_time} seconds")

    def results(self, execution_id: str, max_rows: int = 1000) -> pl.DataFrame:
        """Get query results as a Polars DataFrame using PyArrow for better type handling"""
        try:
            # Get first batch to extract schema
            response = self.client.get_query_results(
                QueryExecutionId=execution_id, MaxResults=min(max_rows, 1000)
            )

            result_set = response["ResultSet"]
            column_info = result_set["ResultSetMetadata"]["ColumnInfo"]

            # Create PyArrow schema
            schema_fields = []
            for col in column_info:
                pa_type = presto_type_to_pyarrow(col["Type"])
                schema_fields.append(pa.field(col["Name"], pa_type))

            schema = pa.schema(schema_fields)

            # Initialize column data collectors
            columns_data = {col["Name"]: [] for col in column_info}

            # Process all batches
            next_token = None
            total_rows = 0

            while total_rows < max_rows:
                if next_token is None:
                    # First batch - skip header row
                    data_rows = result_set["Rows"][1:] if result_set["Rows"] else []
                else:
                    # Get next batch
                    params = {
                        "QueryExecutionId": execution_id,
                        "MaxResults": min(max_rows - total_rows, 1000),
                        "NextToken": next_token,
                    }
                    response = self.client.get_query_results(**params)
                    result_set = response["ResultSet"]
                    data_rows = result_set["Rows"]

                # Extract data from this batch
                for row in data_rows:
                    if total_rows >= max_rows:
                        break

                    for i, col_data in enumerate(row["Data"]):
                        col_name = column_info[i]["Name"]
                        col_type = column_info[i]["Type"]
                        value = self._extract_typed_value(col_data, col_type)
                        columns_data[col_name].append(value)

                    total_rows += 1

                # Check for more results
                next_token = response.get("NextToken")
                if not next_token or total_rows >= max_rows:
                    break

            # Create PyArrow table
            if total_rows == 0:
                # Empty result
                return cast(
                    pl.DataFrame,
                    pl.from_arrow(
                        pa.table(
                            {col["Name"]: [] for col in column_info}, schema=schema
                        )
                    ),
                )

            # Build PyArrow arrays with proper typing
            arrays = []
            for field in schema:
                col_name = field.name
                data = columns_data[col_name]

                try:
                    # Let PyArrow handle the conversion with the specified type
                    array = pa.array(data, type=field.type)
                except (pa.ArrowInvalid, pa.ArrowTypeError):
                    # Fallback to string type if conversion fails
                    array = pa.array(
                        [str(x) if x is not None else None for x in data],
                        type=pa.string(),
                    )

                arrays.append(array)

            # Create table and convert to Polars
            table = pa.table(arrays, names=[field.name for field in schema])
            return cast(pl.DataFrame, pl.from_arrow(table))

        except Exception as e:
            raise Exception(f"Failed to retrieve query results: {e}") from e

    def results_iter(self, execution_id: str, page_size: int = 1000) -> Iterator[pl.DataFrame]:
        """Yield query results page-by-page as Polars DataFrames.

        Args:
            execution_id: The query execution ID.
            page_size: Rows per page (capped at 1000, the Athena API limit).

        Yields:
            One ``pl.DataFrame`` per API page.
        """
        page_size = min(page_size, 1000)

        try:
            # First page – extract schema and skip header row
            response = self.client.get_query_results(
                QueryExecutionId=execution_id, MaxResults=page_size
            )

            result_set = response["ResultSet"]
            column_info = result_set["ResultSetMetadata"]["ColumnInfo"]

            # Build PyArrow schema (reused for every page)
            schema_fields = []
            for col in column_info:
                pa_type = presto_type_to_pyarrow(col["Type"])
                schema_fields.append(pa.field(col["Name"], pa_type))
            schema = pa.schema(schema_fields)

            # First batch skips the header row
            data_rows = result_set["Rows"][1:] if result_set["Rows"] else []

            while True:
                if data_rows:
                    columns_data = {col["Name"]: [] for col in column_info}
                    for row in data_rows:
                        for i, col_data in enumerate(row["Data"]):
                            col_name = column_info[i]["Name"]
                            col_type = column_info[i]["Type"]
                            value = self._extract_typed_value(col_data, col_type)
                            columns_data[col_name].append(value)

                    arrays = []
                    for field in schema:
                        data = columns_data[field.name]
                        try:
                            array = pa.array(data, type=field.type)
                        except (pa.ArrowInvalid, pa.ArrowTypeError):
                            array = pa.array(
                                [str(x) if x is not None else None for x in data],
                                type=pa.string(),
                            )
                        arrays.append(array)

                    table = pa.table(arrays, names=[field.name for field in schema])
                    yield cast(pl.DataFrame, pl.from_arrow(table))

                next_token = response.get("NextToken")
                if not next_token:
                    break

                response = self.client.get_query_results(
                    QueryExecutionId=execution_id,
                    MaxResults=page_size,
                    NextToken=next_token,
                )
                result_set = response["ResultSet"]
                data_rows = result_set["Rows"]

        except Exception as e:
            raise Exception(f"Failed to retrieve query results: {e}") from e

    def _extract_typed_value(self, col_data: dict[str, Any], athena_type: str) -> Any:
        """Extract and convert value based on Athena type"""
        if col_data.get("NullValue"):
            return None

        # Get the raw value
        if "VarCharValue" in col_data:
            raw_value = col_data["VarCharValue"]
        elif "BigIntValue" in col_data:
            return col_data["BigIntValue"]
        elif "DoubleValue" in col_data:
            return col_data["DoubleValue"]
        elif "BooleanValue" in col_data:
            return col_data["BooleanValue"]
        else:
            return None

        # Convert based on type
        athena_type = athena_type.lower()

        # For string-like types, preserve empty strings as-is
        is_string_type = (
            athena_type.startswith("varchar")
            or athena_type.startswith("char")
            or athena_type in ("string", "text")
        )

        if not raw_value and raw_value != "":
            return None

        if not is_string_type and raw_value == "":
            return None

        try:
            if athena_type in ["boolean", "bool"]:
                return raw_value.lower() in ("true", "1", "yes")
            elif athena_type in ["tinyint", "smallint", "int", "integer"]:
                return int(raw_value)
            elif athena_type in ["bigint", "long"]:
                return int(raw_value)
            elif athena_type in ["float", "real", "double", "double precision"]:
                return float(raw_value)
            elif athena_type.startswith("decimal") or athena_type.startswith("numeric"):
                return float(raw_value)  # PyArrow will handle decimal conversion
            else:
                return raw_value  # Keep as string for dates, timestamps, etc.
        except (ValueError, TypeError):
            return raw_value  # Fallback to string if conversion fails

    def get_query_info(self, execution_id: str) -> dict[str, Any]:
        """Get detailed information about a query execution"""
        try:
            response = self.client.get_query_execution(QueryExecutionId=execution_id)
            return response["QueryExecution"]
        except Exception as e:
            raise Exception(f"Failed to get query info: {e}") from e

    def cancel_query(self, execution_id: str) -> bool:
        """Cancel a running query"""
        try:
            self.client.stop_query_execution(QueryExecutionId=execution_id)
            return True
        except Exception as e:
            raise Exception(f"Failed to cancel query: {e}") from e

    def list_work_groups(self) -> list[str]:
        """List available Athena work groups"""
        try:
            response = self.client.list_work_groups()
            return [wg["Name"] for wg in response["WorkGroups"]]
        except Exception as e:
            raise Exception(f"Failed to list work groups: {e}") from e

    @classmethod
    def from_session(
        cls, session: boto3.Session, database: str, output_location: str, **kwargs
    ) -> "AthenaClient":
        """Create AthenaClient from existing boto3 session"""
        return cls(database, output_location, session=session, **kwargs)



"""Tests for AthenaClient"""

import pytest
import boto3
import polars as pl
from unittest.mock import Mock, patch, MagicMock
from botocore.config import Config
from botocore.exceptions import ClientError

from owlbear import AthenaClient


@pytest.fixture
def mock_session():
    """Mock boto3 session"""
    session = Mock(spec=boto3.Session)
    client = Mock()
    session.client.return_value = client
    return session, client


@pytest.fixture
def athena_client():
    """Create AthenaClient with mocked AWS client"""
    with patch('boto3.client') as mock_client:
        client = AthenaClient(
            database="test_db",
            output_location="s3://test-bucket/results/",
            region="us-east-1"
        )
        return client, mock_client.return_value


class TestAthenaClientInit:
    """Test AthenaClient initialization"""

    def test_init_default(self):
        """Test default initialization"""
        with patch('boto3.client') as mock_client:
            client = AthenaClient(
                database="test_db",
                output_location="s3://test-bucket/results/"
            )
            assert client.database == "test_db"
            assert client.output_location == "s3://test-bucket/results/"
            mock_client.assert_called_once()

    def test_init_with_session(self, mock_session):
        """Test initialization with provided session"""
        session, mock_client = mock_session

        client = AthenaClient(
            database="test_db",
            output_location="s3://test-bucket/results/",
            session=session
        )

        assert client.database == "test_db"
        assert client.output_location == "s3://test-bucket/results/"
        session.client.assert_called_once_with('athena', config=None)

    def test_init_with_config(self):
        """Test initialization with custom config"""
        config = Config(region_name='us-west-2')

        with patch('boto3.client') as mock_client:
            client = AthenaClient(
                database="test_db",
                output_location="s3://test-bucket/results/",
                config=config
            )

            mock_client.assert_called_once()
            args, kwargs = mock_client.call_args
            assert kwargs['config'] == config

    def test_from_session(self, mock_session):
        """Test from_session class method"""
        session, mock_client = mock_session

        client = AthenaClient.from_session(
            session=session,
            database="test_db",
            output_location="s3://test-bucket/results/"
        )

        assert client.database == "test_db"
        session.client.assert_called_once()


class TestExecuteQuery:
    """Test query execution"""

    def test_query_basic(self, athena_client):
        """Test basic query execution"""
        client, mock_athena = athena_client
        mock_athena.start_query_execution.return_value = {'QueryExecutionId': 'test-id'}
        mock_athena.get_query_execution.return_value = {
            'QueryExecution': {'Status': {'State': 'SUCCEEDED'}}
        }

        result = client.query("SELECT * FROM test_table")

        assert result == "test-id"
        mock_athena.start_query_execution.assert_called_once()
        call_args = mock_athena.start_query_execution.call_args[1]
        assert call_args['QueryString'] == "SELECT * FROM test_table"
        assert call_args['QueryExecutionContext']['Database'] == "test_db"
        assert call_args['ResultConfiguration']['OutputLocation'] == "s3://test-bucket/results/"

    def test_query_no_wait(self, athena_client):
        """Test query execution without waiting"""
        client, mock_athena = athena_client
        mock_athena.start_query_execution.return_value = {'QueryExecutionId': 'test-id'}

        result = client.query("SELECT * FROM test_table", wait_for_completion=False)

        assert result == "test-id"
        mock_athena.get_query_execution.assert_not_called()

    def test_query_with_workgroup(self, athena_client):
        """Test query execution with work group"""
        client, mock_athena = athena_client
        mock_athena.start_query_execution.return_value = {'QueryExecutionId': 'test-id'}
        mock_athena.get_query_execution.return_value = {
            'QueryExecution': {'Status': {'State': 'SUCCEEDED'}}
        }

        client.query("SELECT * FROM test_table", work_group="test-workgroup")

        call_args = mock_athena.start_query_execution.call_args[1]
        assert call_args['WorkGroup'] == "test-workgroup"

    def test_query_failure(self, athena_client):
        """Test query execution failure"""
        client, mock_athena = athena_client
        mock_athena.start_query_execution.side_effect = Exception("AWS Error")

        with pytest.raises(Exception, match="Failed to execute query"):
            client.query("SELECT * FROM test_table")

    def test_query_with_parameters(self, athena_client):
        """Test parameterized query execution"""
        client, mock_athena = athena_client
        mock_athena.start_query_execution.return_value = {'QueryExecutionId': 'test-id'}
        mock_athena.get_query_execution.return_value = {
            'QueryExecution': {'Status': {'State': 'SUCCEEDED'}}
        }

        client.query(
            "SELECT * FROM t WHERE id = ?",
            parameters=["42"],
        )

        call_args = mock_athena.start_query_execution.call_args[1]
        assert call_args['ExecutionParameters'] == ["42"]

    def test_query_with_result_reuse(self, athena_client):
        """Test query with result reuse configuration"""
        client, mock_athena = athena_client
        mock_athena.start_query_execution.return_value = {'QueryExecutionId': 'test-id'}
        mock_athena.get_query_execution.return_value = {
            'QueryExecution': {'Status': {'State': 'SUCCEEDED'}}
        }

        client.query("SELECT * FROM t", result_reuse_max_age=60)

        call_args = mock_athena.start_query_execution.call_args[1]
        reuse_config = call_args['ResultReuseConfiguration']
        assert reuse_config == {
            'ResultReuseByAgeConfiguration': {
                'Enabled': True,
                'MaxAgeInMinutes': 60,
            }
        }

    def test_query_exception_chaining(self, athena_client):
        """Test that exceptions are chained with 'from e'"""
        client, mock_athena = athena_client
        original = ValueError("original error")
        mock_athena.start_query_execution.side_effect = original

        with pytest.raises(Exception) as exc_info:
            client.query("SELECT 1")

        assert exc_info.value.__cause__ is original


class TestWaitForCompletion:
    """Test _wait_for_completion method"""

    def test_wait_success(self, athena_client):
        """Test successful wait"""
        client, mock_athena = athena_client
        mock_athena.get_query_execution.return_value = {
            'QueryExecution': {'Status': {'State': 'SUCCEEDED'}}
        }

        result = client._wait_for_completion('test-id')
        assert result == 'SUCCEEDED'

    def test_wait_failed_query(self, athena_client):
        """Test wait for failed query"""
        client, mock_athena = athena_client
        mock_athena.get_query_execution.return_value = {
            'QueryExecution': {
                'Status': {
                    'State': 'FAILED',
                    'StateChangeReason': 'Syntax error',
                    'AthenaError': {'ErrorMessage': 'Invalid SQL'}
                }
            }
        }

        with pytest.raises(Exception, match="Query failed: Syntax error. Error: Invalid SQL"):
            client._wait_for_completion('test-id')

    def test_wait_cancelled_query(self, athena_client):
        """Test wait for cancelled query"""
        client, mock_athena = athena_client
        mock_athena.get_query_execution.return_value = {
            'QueryExecution': {'Status': {'State': 'CANCELLED'}}
        }

        with pytest.raises(Exception, match="Query was cancelled"):
            client._wait_for_completion('test-id')

    @patch('time.sleep')
    def test_wait_timeout(self, mock_sleep, athena_client):
        """Test wait timeout"""
        client, mock_athena = athena_client
        mock_athena.get_query_execution.return_value = {
            'QueryExecution': {'Status': {'State': 'RUNNING'}}
        }

        with pytest.raises(TimeoutError, match="Query did not complete within 5 seconds"):
            client._wait_for_completion('test-id', max_wait_time=5)


class TestGetResultsPolars:
    """Test results method"""

    def test_get_results_basic(self, athena_client):
        """Test basic results retrieval with PyArrow type handling"""
        client, mock_athena = athena_client

        # Mock response with typed data
        mock_athena.get_query_results.return_value = {
            'ResultSet': {
                'ResultSetMetadata': {
                    'ColumnInfo': [
                        {'Name': 'id', 'Type': 'integer'},
                        {'Name': 'name', 'Type': 'varchar'},
                        {'Name': 'age', 'Type': 'integer'}
                    ]
                },
                'Rows': [
                    # Header row (skipped)
                    {'Data': [{'VarCharValue': 'id'}, {'VarCharValue': 'name'}, {'VarCharValue': 'age'}]},
                    # Data rows
                    {'Data': [{'VarCharValue': '1'}, {'VarCharValue': 'Alice'}, {'VarCharValue': '25'}]},
                    {'Data': [{'VarCharValue': '2'}, {'VarCharValue': 'Bob'}, {'VarCharValue': '30'}]}
                ]
            }
        }

        df = client.results('test-id')

        assert isinstance(df, pl.DataFrame)
        assert df.columns == ['id', 'name', 'age']
        assert len(df) == 2

        # Check that types are properly inferred
        assert df.dtypes[0] == pl.Int32  # id should be integer
        assert df.dtypes[1] == pl.Utf8   # name should be string
        assert df.dtypes[2] == pl.Int32  # age should be integer

        # Check actual values
        assert df.item(0, 'id') == 1
        assert df.item(0, 'name') == 'Alice'
        assert df.item(0, 'age') == 25

    def test_get_results_empty(self, athena_client):
        """Test empty results"""
        client, mock_athena = athena_client

        mock_athena.get_query_results.return_value = {
            'ResultSet': {
                'ResultSetMetadata': {
                    'ColumnInfo': [
                        {'Name': 'id', 'Type': 'integer'},
                        {'Name': 'name', 'Type': 'varchar'}
                    ]
                },
                'Rows': [
                    # Only header row
                    {'Data': [{'VarCharValue': 'id'}, {'VarCharValue': 'name'}]}
                ]
            }
        }

        df = client.results('test-id')

        assert isinstance(df, pl.DataFrame)
        assert df.columns == ['id', 'name']
        assert len(df) == 0

    def test_get_results_with_nulls(self, athena_client):
        """Test results with null values"""
        client, mock_athena = athena_client

        mock_athena.get_query_results.return_value = {
            'ResultSet': {
                'ResultSetMetadata': {
                    'ColumnInfo': [
                        {'Name': 'id', 'Type': 'integer'},
                        {'Name': 'value', 'Type': 'varchar'}
                    ]
                },
                'Rows': [
                    {'Data': [{'VarCharValue': 'id'}, {'VarCharValue': 'value'}]},
                    {'Data': [{'VarCharValue': '1'}, {'NullValue': True}]},
                    {'Data': [{'VarCharValue': '2'}, {'VarCharValue': 'test'}]}
                ]
            }
        }

        df = client.results('test-id')

        assert len(df) == 2
        assert df.item(0, 'id') == 1  # Should be converted to integer
        assert df.item(0, 'value') is None  # Should handle null
        assert df.item(1, 'value') == 'test'

    def test_get_results_pagination(self, athena_client):
        """Test results with pagination"""
        client, mock_athena = athena_client

        # First page
        first_response = {
            'ResultSet': {
                'ResultSetMetadata': {
                    'ColumnInfo': [{'Name': 'id', 'Type': 'integer'}]
                },
                'Rows': [
                    {'Data': [{'VarCharValue': 'id'}]},
                    {'Data': [{'VarCharValue': '1'}]}
                ]
            },
            'NextToken': 'token123'
        }

        # Second page
        second_response = {
            'ResultSet': {
                'Rows': [
                    {'Data': [{'VarCharValue': '2'}]}
                ]
            }
        }

        mock_athena.get_query_results.side_effect = [first_response, second_response]

        df = client.results('test-id')

        assert len(df) == 2
        assert df.item(0, 'id') == 1  # Should be converted to integer
        assert df.item(1, 'id') == 2
        assert mock_athena.get_query_results.call_count == 2

    def test_get_results_error(self, athena_client):
        """Test results retrieval error"""
        client, mock_athena = athena_client
        mock_athena.get_query_results.side_effect = Exception("AWS Error")

        with pytest.raises(Exception, match="Failed to retrieve query results"):
            client.results('test-id')

    def test_results_exception_chaining(self, athena_client):
        """Test that results exceptions are chained"""
        client, mock_athena = athena_client
        original = RuntimeError("network error")
        mock_athena.get_query_results.side_effect = original

        with pytest.raises(Exception) as exc_info:
            client.results('test-id')

        assert exc_info.value.__cause__ is original


class TestExtractTypedValue:
    """Test _extract_typed_value preserves empty strings for string types"""

    def test_empty_string_preserved_for_varchar(self, athena_client):
        client, _ = athena_client
        result = client._extract_typed_value({'VarCharValue': ''}, 'varchar')
        assert result == ''

    def test_empty_string_preserved_for_string(self, athena_client):
        client, _ = athena_client
        result = client._extract_typed_value({'VarCharValue': ''}, 'string')
        assert result == ''

    def test_empty_string_preserved_for_char(self, athena_client):
        client, _ = athena_client
        result = client._extract_typed_value({'VarCharValue': ''}, 'char(10)')
        assert result == ''

    def test_empty_string_null_for_integer(self, athena_client):
        client, _ = athena_client
        result = client._extract_typed_value({'VarCharValue': ''}, 'integer')
        assert result is None

    def test_empty_string_null_for_double(self, athena_client):
        client, _ = athena_client
        result = client._extract_typed_value({'VarCharValue': ''}, 'double')
        assert result is None


class TestResultsIter:
    """Test results_iter method"""

    def test_single_page(self, athena_client):
        """Single page, no NextToken — yields one DataFrame"""
        client, mock_athena = athena_client

        mock_athena.get_query_results.return_value = {
            'ResultSet': {
                'ResultSetMetadata': {
                    'ColumnInfo': [
                        {'Name': 'id', 'Type': 'integer'},
                        {'Name': 'name', 'Type': 'varchar'},
                    ]
                },
                'Rows': [
                    {'Data': [{'VarCharValue': 'id'}, {'VarCharValue': 'name'}]},
                    {'Data': [{'VarCharValue': '1'}, {'VarCharValue': 'Alice'}]},
                    {'Data': [{'VarCharValue': '2'}, {'VarCharValue': 'Bob'}]},
                ]
            }
        }

        pages = list(client.results_iter('test-id'))

        assert len(pages) == 1
        df = pages[0]
        assert df.columns == ['id', 'name']
        assert len(df) == 2
        assert df.item(0, 'id') == 1
        assert df.item(1, 'name') == 'Bob'

    def test_multi_page(self, athena_client):
        """Multiple pages via NextToken"""
        client, mock_athena = athena_client

        first_response = {
            'ResultSet': {
                'ResultSetMetadata': {
                    'ColumnInfo': [{'Name': 'id', 'Type': 'integer'}]
                },
                'Rows': [
                    {'Data': [{'VarCharValue': 'id'}]},
                    {'Data': [{'VarCharValue': '1'}]},
                ]
            },
            'NextToken': 'token123',
        }
        second_response = {
            'ResultSet': {
                'Rows': [
                    {'Data': [{'VarCharValue': '2'}]},
                ]
            },
        }

        mock_athena.get_query_results.side_effect = [first_response, second_response]

        pages = list(client.results_iter('test-id'))

        assert len(pages) == 2
        assert pages[0].item(0, 'id') == 1
        assert pages[1].item(0, 'id') == 2
        assert mock_athena.get_query_results.call_count == 2

    def test_page_size_capped_at_1000(self, athena_client):
        """page_size > 1000 is capped to 1000"""
        client, mock_athena = athena_client

        mock_athena.get_query_results.return_value = {
            'ResultSet': {
                'ResultSetMetadata': {
                    'ColumnInfo': [{'Name': 'id', 'Type': 'integer'}]
                },
                'Rows': [
                    {'Data': [{'VarCharValue': 'id'}]},
                    {'Data': [{'VarCharValue': '1'}]},
                ]
            }
        }

        list(client.results_iter('test-id', page_size=5000))

        call_args = mock_athena.get_query_results.call_args[1]
        assert call_args['MaxResults'] == 1000

    def test_empty_results(self, athena_client):
        """Only header row — yields nothing"""
        client, mock_athena = athena_client

        mock_athena.get_query_results.return_value = {
            'ResultSet': {
                'ResultSetMetadata': {
                    'ColumnInfo': [{'Name': 'id', 'Type': 'integer'}]
                },
                'Rows': [
                    {'Data': [{'VarCharValue': 'id'}]},
                ]
            }
        }

        pages = list(client.results_iter('test-id'))
        assert len(pages) == 0

    def test_error_handling(self, athena_client):
        """Errors are wrapped consistently"""
        client, mock_athena = athena_client
        mock_athena.get_query_results.side_effect = Exception("AWS Error")

        with pytest.raises(Exception, match="Failed to retrieve query results"):
            list(client.results_iter('test-id'))


class TestQueryInfo:
    """Test get_query_info method"""

    def test_get_query_info(self, athena_client):
        """Test query info retrieval"""
        client, mock_athena = athena_client

        expected_info = {
            'QueryExecutionId': 'test-id',
            'Query': 'SELECT * FROM test',
            'Status': {'State': 'SUCCEEDED'}
        }
        mock_athena.get_query_execution.return_value = {'QueryExecution': expected_info}

        info = client.get_query_info('test-id')
        assert info == expected_info

    def test_get_query_info_error(self, athena_client):
        """Test query info error"""
        client, mock_athena = athena_client
        mock_athena.get_query_execution.side_effect = Exception("AWS Error")

        with pytest.raises(Exception, match="Failed to get query info"):
            client.get_query_info('test-id')


class TestCancelQuery:
    """Test cancel_query method"""

    def test_cancel_query(self, athena_client):
        """Test query cancellation"""
        client, mock_athena = athena_client

        result = client.cancel_query('test-id')
        assert result is True
        mock_athena.stop_query_execution.assert_called_once_with(QueryExecutionId='test-id')

    def test_cancel_query_error(self, athena_client):
        """Test query cancellation error"""
        client, mock_athena = athena_client
        mock_athena.stop_query_execution.side_effect = Exception("AWS Error")

        with pytest.raises(Exception, match="Failed to cancel query"):
            client.cancel_query('test-id')


class TestListWorkGroups:
    """Test list_work_groups method"""

    def test_list_work_groups(self, athena_client):
        """Test work groups listing"""
        client, mock_athena = athena_client

        mock_athena.list_work_groups.return_value = {
            'WorkGroups': [
                {'Name': 'primary'},
                {'Name': 'analytics'},
                {'Name': 'test'}
            ]
        }

        groups = client.list_work_groups()
        assert groups == ['primary', 'analytics', 'test']

    def test_list_work_groups_error(self, athena_client):
        """Test work groups listing error"""
        client, mock_athena = athena_client
        mock_athena.list_work_groups.side_effect = Exception("AWS Error")

        with pytest.raises(Exception, match="Failed to list work groups"):
            client.list_work_groups()



"""Tests for presto_type_to_pyarrow type conversion."""

import pytest
import pyarrow as pa

from owlbear.types import presto_type_to_pyarrow


class TestBasicTypes:
    def test_boolean(self):
        assert presto_type_to_pyarrow("boolean") == pa.bool_()
        assert presto_type_to_pyarrow("bool") == pa.bool_()
        assert presto_type_to_pyarrow("BOOLEAN") == pa.bool_()

    def test_tinyint(self):
        assert presto_type_to_pyarrow("tinyint") == pa.int8()

    def test_smallint(self):
        assert presto_type_to_pyarrow("smallint") == pa.int16()

    def test_integer(self):
        assert presto_type_to_pyarrow("int") == pa.int32()
        assert presto_type_to_pyarrow("integer") == pa.int32()
        assert presto_type_to_pyarrow("INTEGER") == pa.int32()

    def test_bigint(self):
        assert presto_type_to_pyarrow("bigint") == pa.int64()
        assert presto_type_to_pyarrow("long") == pa.int64()

    def test_float(self):
        assert presto_type_to_pyarrow("float") == pa.float32()
        assert presto_type_to_pyarrow("real") == pa.float32()

    def test_double(self):
        assert presto_type_to_pyarrow("double") == pa.float64()
        assert presto_type_to_pyarrow("double precision") == pa.float64()

    def test_date(self):
        assert presto_type_to_pyarrow("date") == pa.date32()

    def test_timestamp(self):
        assert presto_type_to_pyarrow("timestamp") == pa.timestamp("us")

    def test_timestamp_with_time_zone(self):
        assert presto_type_to_pyarrow("timestamp with time zone") == pa.timestamp("us", tz="UTC")

    def test_time(self):
        assert presto_type_to_pyarrow("time") == pa.time64("us")
        assert presto_type_to_pyarrow("time with time zone") == pa.time64("us")

    def test_interval(self):
        assert presto_type_to_pyarrow("interval day to second") == pa.duration("us")
        assert presto_type_to_pyarrow("interval year to month") == pa.month_day_nano_interval()


class TestDecimalTypes:
    def test_decimal_with_params(self):
        assert presto_type_to_pyarrow("decimal(10,2)") == pa.decimal128(10, 2)

    def test_decimal_precision_only(self):
        assert presto_type_to_pyarrow("decimal(10)") == pa.decimal128(10, 0)

    def test_decimal_default(self):
        assert presto_type_to_pyarrow("decimal") == pa.decimal128(38, 18)

    def test_numeric(self):
        assert presto_type_to_pyarrow("numeric(5,3)") == pa.decimal128(5, 3)


class TestStringTypes:
    def test_varchar(self):
        assert presto_type_to_pyarrow("varchar") == pa.string()
        assert presto_type_to_pyarrow("varchar(255)") == pa.string()

    def test_char(self):
        assert presto_type_to_pyarrow("char") == pa.string()
        assert presto_type_to_pyarrow("char(10)") == pa.string()

    def test_string(self):
        assert presto_type_to_pyarrow("string") == pa.string()

    def test_text(self):
        assert presto_type_to_pyarrow("text") == pa.string()


class TestBinaryTypes:
    def test_varbinary(self):
        assert presto_type_to_pyarrow("varbinary") == pa.binary()

    def test_binary(self):
        assert presto_type_to_pyarrow("binary") == pa.binary()


class TestComplexTypes:
    def test_array_of_integers(self):
        result = presto_type_to_pyarrow("array<integer>")
        assert result == pa.list_(pa.int32())

    def test_array_of_strings(self):
        result = presto_type_to_pyarrow("array<varchar>")
        assert result == pa.list_(pa.string())

    def test_nested_array(self):
        result = presto_type_to_pyarrow("array<array<integer>>")
        assert result == pa.list_(pa.list_(pa.int32()))

    def test_map_string_string(self):
        result = presto_type_to_pyarrow("map<varchar,varchar>")
        assert result == pa.map_(pa.string(), pa.string())

    def test_map_string_bigint(self):
        result = presto_type_to_pyarrow("map<varchar,bigint>")
        assert result == pa.map_(pa.string(), pa.int64())

    def test_map_with_complex_value(self):
        result = presto_type_to_pyarrow("map<varchar,array<integer>>")
        assert result == pa.map_(pa.string(), pa.list_(pa.int32()))

    def test_array_of_map(self):
        result = presto_type_to_pyarrow("array<map<varchar,integer>>")
        assert result == pa.list_(pa.map_(pa.string(), pa.int32()))


class TestUnknownTypes:
    def test_unknown_defaults_to_string(self):
        assert presto_type_to_pyarrow("json") == pa.string()
        assert presto_type_to_pyarrow("ipaddress") == pa.string()



"""Tests for TrinoClient"""

import pytest
import polars as pl
from unittest.mock import Mock, patch, MagicMock

from owlbear import TrinoClient


@pytest.fixture
def mock_cursor():
    """Create a mock Trino cursor with description and data."""
    cursor = Mock()
    return cursor


@pytest.fixture
def mock_connection(mock_cursor):
    """Create a mock Trino connection."""
    conn = Mock()
    conn.cursor.return_value = mock_cursor
    return conn


@pytest.fixture
def trino_client():
    """Create a TrinoClient instance (no real connection needed)."""
    return TrinoClient(
        host="trino.example.com",
        port=443,
        user="test_user",
        catalog="hive",
        schema="default",
    )


class TestTrinoClientInit:
    def test_init_defaults(self):
        client = TrinoClient(host="localhost")
        assert client.host == "localhost"
        assert client.port == 443
        assert client.http_scheme == "https"
        assert client.user is None
        assert client.catalog is None
        assert client.schema is None

    def test_init_all_params(self):
        auth = Mock()
        client = TrinoClient(
            host="trino.example.com",
            port=8080,
            user="admin",
            catalog="hive",
            schema="analytics",
            auth=auth,
            http_scheme="http",
        )
        assert client.host == "trino.example.com"
        assert client.port == 8080
        assert client.user == "admin"
        assert client.catalog == "hive"
        assert client.schema == "analytics"
        assert client.auth is auth
        assert client.http_scheme == "http"

    def test_init_extra_kwargs(self):
        client = TrinoClient(host="localhost", source="owlbear-test")
        assert client.kwargs == {"source": "owlbear-test"}


class TestTrinoExecuteQuery:
    def test_basic_query(self, trino_client, mock_connection, mock_cursor):
        mock_cursor.description = [
            ("id", "integer", None, None, None, None, None),
            ("name", "varchar", None, None, None, None, None),
        ]
        mock_cursor.fetchall.return_value = [
            (1, "Alice"),
            (2, "Bob"),
        ]

        with patch.object(trino_client, "_get_connection", return_value=mock_connection):
            df = trino_client.query("SELECT id, name FROM users")

        assert isinstance(df, pl.DataFrame)
        assert df.columns == ["id", "name"]
        assert len(df) == 2
        assert df.item(0, "id") == 1
        assert df.item(0, "name") == "Alice"
        assert df.item(1, "id") == 2
        assert df.item(1, "name") == "Bob"
        mock_connection.close.assert_called_once()

    def test_typed_columns(self, trino_client, mock_connection, mock_cursor):
        mock_cursor.description = [
            ("count", "bigint", None, None, None, None, None),
            ("ratio", "double", None, None, None, None, None),
            ("active", "boolean", None, None, None, None, None),
        ]
        mock_cursor.fetchall.return_value = [
            (100, 0.75, True),
            (200, 0.50, False),
        ]

        with patch.object(trino_client, "_get_connection", return_value=mock_connection):
            df = trino_client.query("SELECT count, ratio, active FROM stats")

        assert df.dtypes[0] == pl.Int64
        assert df.dtypes[1] == pl.Float64
        assert df.dtypes[2] == pl.Boolean

    def test_with_nulls(self, trino_client, mock_connection, mock_cursor):
        mock_cursor.description = [
            ("id", "integer", None, None, None, None, None),
            ("value", "varchar", None, None, None, None, None),
        ]
        mock_cursor.fetchall.return_value = [
            (1, None),
            (2, "test"),
        ]

        with patch.object(trino_client, "_get_connection", return_value=mock_connection):
            df = trino_client.query("SELECT id, value FROM t")

        assert len(df) == 2
        assert df.item(0, "value") is None
        assert df.item(1, "value") == "test"

    def test_empty_results(self, trino_client, mock_connection, mock_cursor):
        mock_cursor.description = [
            ("id", "integer", None, None, None, None, None),
            ("name", "varchar", None, None, None, None, None),
        ]
        mock_cursor.fetchall.return_value = []

        with patch.object(trino_client, "_get_connection", return_value=mock_connection):
            df = trino_client.query("SELECT id, name FROM empty_table")

        assert isinstance(df, pl.DataFrame)
        assert df.columns == ["id", "name"]
        assert len(df) == 0

    def test_no_description(self, trino_client, mock_connection, mock_cursor):
        """DDL statements return no description."""
        mock_cursor.description = None
        mock_cursor.fetchall.return_value = []

        with patch.object(trino_client, "_get_connection", return_value=mock_connection):
            df = trino_client.query("CREATE TABLE foo (id INT)")

        assert isinstance(df, pl.DataFrame)
        assert len(df) == 0

    def test_max_rows(self, trino_client, mock_connection, mock_cursor):
        mock_cursor.description = [
            ("id", "integer", None, None, None, None, None),
        ]
        mock_cursor.fetchmany.return_value = [(1,), (2,)]

        with patch.object(trino_client, "_get_connection", return_value=mock_connection):
            df = trino_client.query("SELECT id FROM t", max_rows=2)

        mock_cursor.fetchmany.assert_called_once_with(2)
        assert len(df) == 2

    def test_connection_closed_on_error(self, trino_client, mock_connection, mock_cursor):
        mock_cursor.execute.side_effect = Exception("Connection failed")

        with patch.object(trino_client, "_get_connection", return_value=mock_connection):
            with pytest.raises(Exception, match="Connection failed"):
                trino_client.query("SELECT 1")

        mock_connection.close.assert_called_once()

    def test_query_with_parameters(self, trino_client, mock_connection, mock_cursor):
        """Test parameterized query execution"""
        mock_cursor.description = [
            ("id", "integer", None, None, None, None, None),
        ]
        mock_cursor.fetchall.return_value = [(42,)]

        with patch.object(trino_client, "_get_connection", return_value=mock_connection):
            df = trino_client.query(
                "SELECT id FROM t WHERE id = ?",
                parameters=[42],
            )

        mock_cursor.execute.assert_called_once_with(
            "SELECT id FROM t WHERE id = ?", params=[42]
        )
        assert len(df) == 1
        assert df.item(0, "id") == 42


class TestQueryIter:
    """Test query_iter method"""

    def test_single_batch(self, trino_client, mock_connection, mock_cursor):
        """One batch of rows then exhausted"""
        mock_cursor.description = [
            ("id", "integer", None, None, None, None, None),
            ("name", "varchar", None, None, None, None, None),
        ]
        mock_cursor.fetchmany.side_effect = [
            [(1, "Alice"), (2, "Bob")],
            [],
        ]

        with patch.object(trino_client, "_get_connection", return_value=mock_connection):
            pages = list(trino_client.query_iter("SELECT id, name FROM users"))

        assert len(pages) == 1
        df = pages[0]
        assert df.columns == ["id", "name"]
        assert len(df) == 2
        assert df.item(0, "id") == 1
        assert df.item(1, "name") == "Bob"
        mock_connection.close.assert_called_once()

    def test_multi_batch(self, trino_client, mock_connection, mock_cursor):
        """Multiple batches via fetchmany"""
        mock_cursor.description = [
            ("id", "integer", None, None, None, None, None),
        ]
        mock_cursor.fetchmany.side_effect = [
            [(1,), (2,)],
            [(3,)],
            [],
        ]

        with patch.object(trino_client, "_get_connection", return_value=mock_connection):
            pages = list(trino_client.query_iter("SELECT id FROM t", page_size=2))

        assert len(pages) == 2
        assert pages[0].item(0, "id") == 1
        assert pages[0].item(1, "id") == 2
        assert pages[1].item(0, "id") == 3

    def test_empty_results(self, trino_client, mock_connection, mock_cursor):
        """fetchmany returns empty on first call"""
        mock_cursor.description = [
            ("id", "integer", None, None, None, None, None),
        ]
        mock_cursor.fetchmany.return_value = []

        with patch.object(trino_client, "_get_connection", return_value=mock_connection):
            pages = list(trino_client.query_iter("SELECT id FROM empty_table"))

        assert len(pages) == 0
        mock_connection.close.assert_called_once()

    def test_no_description(self, trino_client, mock_connection, mock_cursor):
        """DDL with no description yields nothing"""
        mock_cursor.description = None

        with patch.object(trino_client, "_get_connection", return_value=mock_connection):
            pages = list(trino_client.query_iter("CREATE TABLE foo (id INT)"))

        assert len(pages) == 0
        mock_connection.close.assert_called_once()

    def test_connection_closed_on_error(self, trino_client, mock_connection, mock_cursor):
        """Connection is closed even when execute raises"""
        mock_cursor.execute.side_effect = Exception("Connection failed")

        with patch.object(trino_client, "_get_connection", return_value=mock_connection):
            with pytest.raises(Exception, match="Connection failed"):
                list(trino_client.query_iter("SELECT 1"))

        mock_connection.close.assert_called_once()

    def test_query_with_parameters(self, trino_client, mock_connection, mock_cursor):
        """Parameters are forwarded to cursor.execute"""
        mock_cursor.description = [
            ("id", "integer", None, None, None, None, None),
        ]
        mock_cursor.fetchmany.side_effect = [[(42,)], []]

        with patch.object(trino_client, "_get_connection", return_value=mock_connection):
            pages = list(trino_client.query_iter(
                "SELECT id FROM t WHERE id = ?",
                parameters=[42],
            ))

        mock_cursor.execute.assert_called_once_with(
            "SELECT id FROM t WHERE id = ?", params=[42]
        )
        assert len(pages) == 1
        assert pages[0].item(0, "id") == 42


class TestTrinoGetConnection:
    def test_get_connection_params(self):
        auth = Mock()
        client = TrinoClient(
            host="trino.example.com",
            port=8080,
            user="admin",
            catalog="hive",
            schema="analytics",
            auth=auth,
            http_scheme="http",
        )

        with patch("trino.dbapi.connect") as mock_connect:
            client._get_connection()
            mock_connect.assert_called_once_with(
                host="trino.example.com",
                port=8080,
                http_scheme="http",
                user="admin",
                catalog="hive",
                schema="analytics",
                auth=auth,
            )

    def test_get_connection_minimal(self):
        client = TrinoClient(host="localhost")

        with patch("trino.dbapi.connect") as mock_connect:
            client._get_connection()
            mock_connect.assert_called_once_with(
                host="localhost",
                port=443,
                http_scheme="https",
            )



"""Shared Presto-family type conversion for Athena and Trino backends."""

import pyarrow as pa


def _find_matching_bracket(s: str, start: int) -> int:
    """Find the index of the '>' that closes the '<' at position start."""
    depth = 0
    for i in range(start, len(s)):
        if s[i] == "<":
            depth += 1
        elif s[i] == ">":
            depth -= 1
            if depth == 0:
                return i
    raise ValueError(f"Unmatched '<' in type string: {s}")


def _split_top_level_comma(s: str) -> list[str]:
    """Split a string on commas that are not nested inside angle brackets."""
    parts = []
    depth = 0
    current: list[str] = []
    for ch in s:
        if ch == "<":
            depth += 1
            current.append(ch)
        elif ch == ">":
            depth -= 1
            current.append(ch)
        elif ch == "," and depth == 0:
            parts.append("".join(current).strip())
            current = []
        else:
            current.append(ch)
    parts.append("".join(current).strip())
    return parts


def presto_type_to_pyarrow(presto_type: str) -> pa.DataType:
    """Convert a Presto/Trino/Athena type name to a PyArrow data type."""
    presto_type = presto_type.lower().strip()

    if presto_type in ["boolean", "bool"]:
        return pa.bool_()
    elif presto_type == "tinyint":
        return pa.int8()
    elif presto_type == "smallint":
        return pa.int16()
    elif presto_type in ["int", "integer"]:
        return pa.int32()
    elif presto_type in ["bigint", "long"]:
        return pa.int64()
    elif presto_type in ["float", "real"]:
        return pa.float32()
    elif presto_type in ["double", "double precision"]:
        return pa.float64()
    elif presto_type == "date":
        return pa.date32()
    elif presto_type in ["time", "time with time zone"]:
        return pa.time64("us")
    elif presto_type == "timestamp with time zone":
        return pa.timestamp("us", tz="UTC")
    elif presto_type.startswith("timestamp"):
        return pa.timestamp("us")
    elif presto_type == "interval day to second":
        return pa.duration("us")
    elif presto_type == "interval year to month":
        return pa.month_day_nano_interval()
    elif presto_type.startswith("decimal") or presto_type.startswith("numeric"):
        if "(" in presto_type:
            params = presto_type.split("(")[1].split(")")[0].split(",")
            precision = int(params[0])
            scale = int(params[1]) if len(params) > 1 else 0
            return pa.decimal128(precision, scale)
        return pa.decimal128(38, 18)
    elif presto_type.startswith("varchar") or presto_type.startswith("char"):
        return pa.string()
    elif presto_type in ["string", "text"]:
        return pa.string()
    elif presto_type in ["varbinary", "binary"]:
        return pa.binary()
    elif presto_type.startswith("array"):
        open_bracket = presto_type.index("<")
        close_bracket = _find_matching_bracket(presto_type, open_bracket)
        element_type_str = presto_type[open_bracket + 1 : close_bracket]
        element_type = presto_type_to_pyarrow(element_type_str)
        return pa.list_(element_type)
    elif presto_type.startswith("map"):
        open_bracket = presto_type.index("<")
        close_bracket = _find_matching_bracket(presto_type, open_bracket)
        inner = presto_type[open_bracket + 1 : close_bracket]
        parts = _split_top_level_comma(inner)
        key_type = presto_type_to_pyarrow(parts[0])
        value_type = presto_type_to_pyarrow(parts[1])
        return pa.map_(key_type, value_type)
    else:
        return pa.string()



"""Owlbear MCP server — expose data-lake queries to AI assistants."""

from __future__ import annotations

import json
import os
import re
from typing import Union

from mcp.server.fastmcp import FastMCP

from .athena import AthenaClient
from .trino import TrinoClient

mcp = FastMCP("owlbear")

_MAX_ROWS_CAP = 10_000

# ---------------------------------------------------------------------------
# Client factory — lazy singleton from env vars
# ---------------------------------------------------------------------------

_client: Union[AthenaClient, TrinoClient, None] = None


def _get_client() -> Union[AthenaClient, TrinoClient]:
    global _client
    if _client is not None:
        return _client

    backend = os.environ.get("OWLBEAR_BACKEND", "athena").lower()

    if backend == "athena":
        _client = AthenaClient(
            database=os.environ.get("OWLBEAR_DATABASE", "default"),
            output_location=os.environ["OWLBEAR_S3_OUTPUT_LOCATION"],
            region=os.environ.get("AWS_REGION", "us-east-1"),
        )
    elif backend == "trino":
        _client = TrinoClient(
            host=os.environ["OWLBEAR_TRINO_HOST"],
            port=int(os.environ.get("OWLBEAR_TRINO_PORT", "443")),
            user=os.environ.get("OWLBEAR_TRINO_USER"),
            catalog=os.environ.get("OWLBEAR_TRINO_CATALOG"),
            schema=os.environ.get("OWLBEAR_DATABASE"),
        )
    else:
        raise ValueError(f"Unknown OWLBEAR_BACKEND: {backend!r} (expected 'athena' or 'trino')")

    return _client


# ---------------------------------------------------------------------------
# Shared helpers
# ---------------------------------------------------------------------------


def _query_to_json(sql: str, max_rows: int) -> str:
    """Run *sql* via the active backend and return JSON-serialised rows."""
    max_rows = min(max_rows, _MAX_ROWS_CAP)
    client = _get_client()

    if isinstance(client, AthenaClient):
        execution_id = client.query(sql)
        df = client.results(execution_id, max_rows=max_rows)
    else:
        df = client.query(sql, max_rows=max_rows)

    return json.dumps(df.to_dicts(), default=str)


_SCALAR_STAT_PATTERN = re.compile(
    r"^("
    r"boolean|tinyint|smallint|int(eger)?|bigint"
    r"|float|double|real"
    r"|varchar|char|string"
    r"|date|timestamp"
    r"|decimal"
    r")",
    re.IGNORECASE,
)


def _is_scalar_stat_type(data_type: str) -> bool:
    """Return True if *data_type* supports MIN/MAX/COUNT(DISTINCT)."""
    return _SCALAR_STAT_PATTERN.match(data_type.strip()) is not None


# ---------------------------------------------------------------------------
# MCP tools
# ---------------------------------------------------------------------------


@mcp.tool()
def execute_query(sql: str, max_rows: int = 500) -> str:
    """Execute a SQL query against the data lake and return results as JSON.

    Args:
        sql: The SQL query to execute.
        max_rows: Maximum rows to return (default 500, capped at 10 000).
    """
    try:
        return _query_to_json(sql, max_rows)
    except Exception as e:
        return json.dumps({"error": str(e)})


@mcp.tool()
def list_databases() -> str:
    """List all databases available in the data lake."""
    try:
        return _query_to_json("SHOW DATABASES", max_rows=_MAX_ROWS_CAP)
    except Exception as e:
        return json.dumps({"error": str(e)})


@mcp.tool()
def list_tables(database: str | None = None) -> str:
    """List tables in a database (defaults to the configured database).

    Args:
        database: Optional database name. Uses the configured default if omitted.
    """
    try:
        sql = f"SHOW TABLES IN {database}" if database else "SHOW TABLES"
        return _query_to_json(sql, max_rows=_MAX_ROWS_CAP)
    except Exception as e:
        return json.dumps({"error": str(e)})


@mcp.tool()
def describe_table(table: str) -> str:
    """Describe the columns and types of a table.

    Args:
        table: Fully-qualified or short table name (e.g. ``my_db.my_table``).
    """
    try:
        return _query_to_json(f"DESCRIBE {table}", max_rows=_MAX_ROWS_CAP)
    except Exception as e:
        return json.dumps({"error": str(e)})


@mcp.tool()
def get_schema_context(tables: str) -> str:
    """Batch DESCRIBE for multiple tables. Returns {table: [columns...]}.

    Args:
        tables: Comma-separated table names (e.g. ``db.t1, db.t2``).
    """
    table_list = [t.strip() for t in tables.split(",") if t.strip()]
    result: dict[str, list[dict[str, str]] | str] = {}
    for table in table_list:
        try:
            raw = json.loads(_query_to_json(f"DESCRIBE {table}", max_rows=_MAX_ROWS_CAP))
            result[table] = [
                row for row in raw if not row.get("col_name", "").startswith("#")
            ]
        except Exception as e:
            result[table] = f"error: {e}"
    return json.dumps(result, default=str)


@mcp.tool()
def profile_table(table: str, sample_size: int = 100) -> str:
    """Profile a table: schema, row count, column stats, and sample rows.

    Args:
        table: Fully-qualified or short table name.
        sample_size: Number of sample rows to return (default 100).
    """
    try:
        # 1. Schema
        describe_raw = json.loads(
            _query_to_json(f"DESCRIBE {table}", max_rows=_MAX_ROWS_CAP)
        )
        columns = [
            row for row in describe_raw if not row.get("col_name", "").startswith("#")
        ]

        # 2. Row count
        count_raw = json.loads(_query_to_json(f"SELECT COUNT(*) AS cnt FROM {table}", max_rows=1))
        row_count = count_raw[0]["cnt"] if count_raw else None

        # 3. Column stats
        stat_parts: list[str] = []
        for col in columns:
            name = col.get("col_name", col.get("column_name", ""))
            dtype = col.get("data_type", col.get("type", ""))
            quoted = f'"{name}"'
            stat_parts.append(f"COUNT(CASE WHEN {quoted} IS NULL THEN 1 END) AS \"{name}__null_count\"")
            if _is_scalar_stat_type(dtype):
                stat_parts.append(
                    f"COUNT(DISTINCT {quoted}) AS \"{name}__distinct_count\""
                )
                stat_parts.append(
                    f"CAST(MIN({quoted}) AS VARCHAR) AS \"{name}__min\""
                )
                stat_parts.append(
                    f"CAST(MAX({quoted}) AS VARCHAR) AS \"{name}__max\""
                )

        if stat_parts:
            stats_sql = f"SELECT {', '.join(stat_parts)} FROM {table}"
            stats_raw = json.loads(_query_to_json(stats_sql, max_rows=1))
            stats_row = stats_raw[0] if stats_raw else {}
        else:
            stats_row = {}

        # Reshape into per-column stats
        column_stats: list[dict[str, object]] = []
        for col in columns:
            name = col.get("col_name", col.get("column_name", ""))
            dtype = col.get("data_type", col.get("type", ""))
            entry: dict[str, object] = {
                "column": name,
                "type": dtype,
                "null_count": stats_row.get(f"{name}__null_count"),
            }
            if _is_scalar_stat_type(dtype):
                entry["distinct_count"] = stats_row.get(f"{name}__distinct_count")
                entry["min"] = stats_row.get(f"{name}__min")
                entry["max"] = stats_row.get(f"{name}__max")
            column_stats.append(entry)

        # 4. Sample rows
        sample_raw = json.loads(
            _query_to_json(f"SELECT * FROM {table} LIMIT {sample_size}", max_rows=sample_size)
        )

        return json.dumps(
            {
                "table": table,
                "row_count": row_count,
                "columns": column_stats,
                "sample_rows": sample_raw,
            },
            default=str,
        )
    except Exception as e:
        return json.dumps({"error": str(e)})


@mcp.tool()
def generate_snippet(table: str, operation: str) -> str:
    """Generate a Python code snippet using owlbear + Polars for a table.

    Args:
        table: Fully-qualified or short table name.
        operation: One of ``load``, ``filter``, ``aggregate``, ``join``.
    """
    valid_ops = {"load", "filter", "aggregate", "join"}
    if operation not in valid_ops:
        return json.dumps({"error": f"Unknown operation {operation!r}. Choose from: {', '.join(sorted(valid_ops))}"})

    try:
        describe_raw = json.loads(
            _query_to_json(f"DESCRIBE {table}", max_rows=_MAX_ROWS_CAP)
        )
        columns = [
            row for row in describe_raw if not row.get("col_name", "").startswith("#")
        ]

        col_names = [c.get("col_name", c.get("column_name", "")) for c in columns]
        col_types = [c.get("data_type", c.get("type", "")) for c in columns]

        # Pick representative columns
        numeric_col = next(
            (n for n, t in zip(col_names, col_types) if _is_scalar_stat_type(t) and re.match(r"(int|bigint|float|double|real|decimal|smallint|tinyint)", t, re.I)),
            col_names[0] if col_names else "col",
        )
        string_col = next(
            (n for n, t in zip(col_names, col_types) if re.match(r"(varchar|char|string)", t, re.I)),
            col_names[0] if col_names else "col",
        )

        header = (
            "from owlbear import AthenaClient\n"
            "import polars as pl\n\n"
            "client = AthenaClient(\n"
            '    database="your_database",\n'
            '    output_location="s3://your-bucket/results/",\n'
            ")\n\n"
        )

        if operation == "load":
            snippet = (
                header
                + f'eid = client.query("SELECT * FROM {table} LIMIT 1000")\n'
                + "df = client.results(eid)\n"
                + "print(df.head())\n"
            )
        elif operation == "filter":
            snippet = (
                header
                + f'eid = client.query("SELECT * FROM {table} WHERE \\"{string_col}\\\\" IS NOT NULL LIMIT 1000")\n'
                + "df = client.results(eid)\n"
                + f'filtered = df.filter(pl.col("{string_col}").is_not_null())\n'
                + "print(filtered.head())\n"
            )
        elif operation == "aggregate":
            snippet = (
                header
                + f'eid = client.query("SELECT * FROM {table} LIMIT 10000")\n'
                + "df = client.results(eid)\n"
                + f'result = df.group_by("{string_col}").agg(pl.col("{numeric_col}").mean())\n'
                + "print(result.head())\n"
            )
        else:  # join
            snippet = (
                header
                + f'eid1 = client.query("SELECT * FROM {table} LIMIT 1000")\n'
                + "df1 = client.results(eid1)\n\n"
                + '# Load a second table to join with\n'
                + 'eid2 = client.query("SELECT * FROM other_table LIMIT 1000")\n'
                + "df2 = client.results(eid2)\n\n"
                + f'joined = df1.join(df2, on="{col_names[0] if col_names else "id"}", how="inner")\n'
                + "print(joined.head())\n"
            )

        return json.dumps(
            {
                "table": table,
                "operation": operation,
                "columns": col_names,
                "snippet": snippet,
            },
            default=str,
        )
    except Exception as e:
        return json.dumps({"error": str(e)})


# ---------------------------------------------------------------------------
# MCP prompts
# ---------------------------------------------------------------------------


@mcp.prompt()
def explore_table(table: str) -> list[dict[str, str]]:
    """Guided exploration of a data-lake table.

    Args:
        table: The table to explore.
    """
    return [
        {
            "role": "user",
            "content": (
                f"I'd like to explore the table **{table}**. Please:\n\n"
                f"1. Run `DESCRIBE {table}` to show the schema.\n"
                f"2. Sample a few rows with `SELECT * FROM {table} LIMIT 5`.\n"
                "3. Suggest 3 interesting analytical queries based on the columns.\n"
                "4. Summarize what this table appears to contain and how it could be used."
            ),
        }
    ]


@mcp.prompt()
def build_pipeline(table: str, goal: str) -> list[dict[str, str]]:
    """Generate a data pipeline using owlbear and Polars.

    Args:
        table: The source table.
        goal: What the pipeline should accomplish.
    """
    return [
        {
            "role": "user",
            "content": (
                f"Help me build a data pipeline for **{table}** to accomplish: {goal}\n\n"
                "Start from this boilerplate:\n\n"
                "```python\n"
                "from owlbear import AthenaClient\n"
                "import polars as pl\n\n"
                "client = AthenaClient(\n"
                '    database="your_database",\n'
                '    output_location="s3://your-bucket/results/",\n'
                ")\n\n"
                f'eid = client.query("SELECT * FROM {table} LIMIT 1000")\n'
                "df = client.results(eid)\n"
                "```\n\n"
                "Adapt this code to achieve the goal. Use Polars for transformations."
            ),
        }
    ]


# ---------------------------------------------------------------------------
# MCP resource
# ---------------------------------------------------------------------------


@mcp.resource("owlbear://config")
def get_config() -> str:
    """Expose the current owlbear backend configuration."""
    backend = os.environ.get("OWLBEAR_BACKEND", "athena").lower()
    config: dict[str, object] = {"backend": backend}

    if backend == "athena":
        config["database"] = os.environ.get("OWLBEAR_DATABASE", "default")
        config["region"] = os.environ.get("AWS_REGION", "us-east-1")
        config["s3_output_location"] = os.environ.get("OWLBEAR_S3_OUTPUT_LOCATION", "")
    elif backend == "trino":
        config["host"] = os.environ.get("OWLBEAR_TRINO_HOST", "")
        config["port"] = int(os.environ.get("OWLBEAR_TRINO_PORT", "443"))
        config["user"] = os.environ.get("OWLBEAR_TRINO_USER", "")
        config["catalog"] = os.environ.get("OWLBEAR_TRINO_CATALOG", "")
        config["database"] = os.environ.get("OWLBEAR_DATABASE", "")

    return json.dumps(config)


# ---------------------------------------------------------------------------
# Entry point
# ---------------------------------------------------------------------------


def main() -> None:
    mcp.run(transport="stdio")


if __name__ == "__main__":
    main()
